"""Utility class for converting Song to/from MIDI.

TODO: implement time signature export to MIDI. Currently there are no time signature change
      messages in the output MIDI file.

TODO: refactor to simplify computation of "delta" values in midi export.
"""

from collections import defaultdict
from enum import IntEnum

from mido import Message, MetaMessage, MidiFile, MidiTrack
from pretty_midi import PrettyMIDI, Instrument

from Core.pretty_midi_utils import get_all_time_signatures
from Core.song import Song
from Core.song_elements import TrackType, Note, Event, Measure


_DRUM_CHANNEL = 9


class _EventType(IntEnum):
    """This ordering is intentional, as it is used to sort events in get_events below."""
    NOTE_OFF = 1
    MEASURE_START = 2
    NOTE_ON = 3


class SongMidiConverter():
    """Class providing Song to/from MIDI conversion routines."""

    @staticmethod
    def _turn_off_notes(midi_track, channel, tick, prev_tick, notes_to_turn_off,  # pylint: disable=dangerous-default-value
                        notes_to_ignore=[]):
        """Create add MIDI messages to turn off all given notes except those in the ignore list."""
        for pitch in notes_to_turn_off:
            if pitch not in notes_to_ignore:
                delta = tick - prev_tick
                prev_tick = tick
                midi_track.append(Message('note_off', channel=channel, note=pitch,
                                          velocity=0, time=delta))
        return prev_tick

    @staticmethod
    def export_midi(song, file_name):
        """Creates a midi file representing the current song and saves the result to a file."""
        midi_file = MidiFile(ticks_per_beat=song.ticks_per_beat)

        # TODO: Output key signature changes to track 0.

        for track_idx, track in enumerate(song):
            midi_track = MidiTrack()
            midi_track.name = track.name
            midi_file.tracks.append(midi_track)
            channel = track.channel
            if channel > 15:
                print('Warning: too many channels for song %s. Skipping extra channel #%d.'
                      % (song.name, channel))
                continue
            midi_track.append(Message('program_change', channel=channel, program=track.program,
                                      time=0))
            measure_number = 0
            prev_tick = 0
            tick = 0
            prev_sounding_notes = set()  # contains MIDI numbers of previously-sounding notes
            previous_time_signature = None

            for measure in track:
                tick = measure.start_tick  # Jump to the start of this measure.

                # Write a time signature message if it has changed.
                # Only do this for track 0.
                if track_idx == 0 and measure.time_signature != previous_time_signature:
                    previous_time_signature = measure.time_signature
                    delta = tick - prev_tick
                    prev_tick = tick
                    midi_track.append(MetaMessage(
                        'time_signature',
                        numerator=measure.time_signature.numerator,
                        denominator=measure.time_signature.denominator,
                        time=delta))

                for event in measure:
                    sounding_notes = set()  # Notes that are sounding at this moment.

                    # First send a note off for all previously-sounding notes EXCEPT those
                    # which show up in the current event with the "tie_from_previous" flag set.
                    tied_from_previous = set(n.midi_pitch for n in event if n.tie_from_previous)
                    prev_tick = SongMidiConverter._turn_off_notes(midi_track, channel, tick,
                                                                  prev_tick, prev_sounding_notes,
                                                                  tied_from_previous)

                    for note in event:
                        pitch = note.midi_pitch
                        velocity = note.velocity
                        tie_from_previous = note.tie_from_previous
                        assert pitch >= 0 and pitch < 128
                        assert tie_from_previous or (velocity >= 0 and velocity < 128)

                        if velocity > 0 and not tie_from_previous:
                            # Handle note-on events.
                            delta = tick - prev_tick
                            prev_tick = tick
                            midi_track.append(Message('note_on', channel=channel, note=pitch,
                                                      velocity=note.velocity, time=delta))
                            sounding_notes.add(pitch)
                        elif not tie_from_previous:
                            # Handle explicit note-off events (note with velocity 0)
                            delta = tick - prev_tick
                            prev_tick = tick
                            midi_track.append(Message('note_off', channel=channel, note=pitch,
                                                      velocity=0, time=delta))
                        else:
                            # Handle tie_from_previous.
                            sounding_notes.add(pitch)

                    prev_sounding_notes = sounding_notes
                    tick += event.duration

                # We processed all events in the measure.  If not at the end of the measure yet,
                # turn off all notes.
                if tick < measure.start_tick + measure.get_duration_ticks():
                    prev_tick = SongMidiConverter._turn_off_notes(midi_track, channel, tick,
                                                                  prev_tick, prev_sounding_notes)
                    prev_sounding_notes = set()

                measure_number += 1

            # Turn off all notes at end of track.
            SongMidiConverter._turn_off_notes(midi_track, channel, tick, prev_tick,
                                              prev_sounding_notes)

            # End of track message.
            midi_track.append(MetaMessage('end_of_track'))

        midi_file.save(file_name)

    @staticmethod
    def create_song_from_midi(file_name):
        """This method creates a new Song object by reading a MIDI file, and parsing the tracks,
        and converting them to song.Track objects. Also sets other Song metadata that can be
        derived from the MIDI."""

        # Load MIDI file into PrettyMIDI object
        try:
            midi = PrettyMIDI(file_name)
        except IndexError as e:
            print("Exception: %s" % e)
            print("Error loading MIDI file in PrettyMIDI. Ensure the file has at least 1 track.")
            return Song()

        instrument_and_type_list = []
        for instrument in midi.instruments:
            if instrument.is_drum:
                instrument_and_type_list.append((instrument, TrackType.DRUMS))
            else:
                instrument_and_type_list.append((instrument, TrackType.UNKNOWN))

        return SongMidiConverter.create_song_from_pretty_midi_instruments(midi,
                                                                          instrument_and_type_list)

    @staticmethod
    def _get_measure_number(time, downbeats):
        for i, downbeat in enumerate(downbeats):
            if time >= downbeat:
                return i - 1
        # Not found.
        raise Exception('measure number not found for time %f\nDownbeats: %s' % (time, downbeats))

    @staticmethod
    def _get_events(instrument, downbeats, midi):
        """Return a sorted list of tuples of (tick, _EventType, note). note is optional.
        Notes here are pretty_mid notes, not Song notes."""
        event_list = []
        #note_starts = defaultdict(list)  # key is start tick, key is Note
        #note_ends = defaultdict(list)  # key is end tick, key is Note
        for note in instrument.notes:
            start_tick = midi.time_to_tick(note.start)
            end_tick = midi.time_to_tick(note.end)
            event_list.append((start_tick, _EventType.NOTE_ON, note))
            event_list.append((end_tick, _EventType.NOTE_OFF, note))
        for downbeat in downbeats:
            tick = midi.time_to_tick(downbeat)
            event_list.append((tick, _EventType.MEASURE_START, None))

        # Sort the list first by tick and second by event type. Thus, for events at the same tick,
        # NOTE_OFF events are listed first, then MEASURE_START, and finally NOTE_ON.

        def get_key(event):
            if event[2] is not None:
                return (event[0], event[1], event[2].pitch)
            return (event[0], event[1])

        return sorted(event_list, key=get_key)

    @staticmethod
    def _write_event(delta_duration, current_measure, notes_to_add, sounding_notes):
        event = current_measure.new_event(delta_duration)
        pitches_written = defaultdict(bool)
        for note in notes_to_add:
            event.append_note(note)
            pitches_written[note.midi_pitch] = True
        for pitch in sounding_notes:
            if sounding_notes[pitch]:
                if not pitches_written[pitch]:
                    event.append_note(Note(pitch, tie_from_previous=True))
                    pitches_written[pitch] = True

    #TODO: add metadata (midi_path, h5_path, original key?), add to song Class
    @staticmethod
    def create_song_from_pretty_midi_instruments(midi, instrument_and_type_list, name=''):
        """Create a new Song using the given list of Instrument objects.

        params:
            midi: a PrettyMidi object that owns the instruments to be added. Used for time to tick
                  conversion.
            instrument_and_type_list: a list of tuples of
                                      (pretty_midi.Instrument, song_elements.TrackType)

        returns: a Song object containing the given Instruments as Tracks.
        """
        assert isinstance(midi, PrettyMIDI)

        downbeats = midi.get_downbeats()
        song = Song(name, midi.resolution)

        # Convert each pretty_midi Instrument into a Song track.
        # N.B. This may change the order of tracks relative to the original MIDI file, and it
        # forces each instrument onto a different channel, which might make us run out of MIDI
        # channels.
        channel = 0
        for instrument, track_type in instrument_and_type_list:
            assert isinstance(instrument, Instrument)
            assert isinstance(track_type, TrackType)

            if instrument.is_drum:
                this_channel = _DRUM_CHANNEL
            else:
                this_channel = channel

            track = song.new_track(instrument.name, instrument.program, this_channel, track_type)

            # Generate set of event times with associated events.
            events = SongMidiConverter._get_events(instrument, downbeats, midi)

            # Get the list of time signatures, one for each measure.
            time_signatures = get_all_time_signatures(midi, downbeats)

            # The events have been sorted by tick and by event type, so for each tick, we can
            # easily process all note-off events, then all measure boundaries, and then all note on
            # events. This allows us to chop up long notes into notes that restart at each measure
            # boundary with a "tie-to-previous" marker. Similarly, all notes are split (with a
            # tie-to previous marker) any time other notes end. The effect is that there is a new
            # Song event at any point that the set of notes changes or we cross a measure boundary.
            #   We keep track of all currently-sounding notes so that we can generate any required
            # tie-to-previous notes at each event time point.

            # key is the midi pitch; value is True if sounding
            sounding_notes = defaultdict(bool)
            prev_tick = 0
            measure_idx = -1
            current_measure = None
            notes_to_add = []  # list of song_elements.Note objects

            first_note_off_at_tick = True
            for tick, event_type, note in events:  # note is a pretty_midi note
                if tick > prev_tick:
                    # We have moved on to a new tick. Generate the event in the previous measure.
                    delta_duration = tick - prev_tick
                    SongMidiConverter._write_event(delta_duration, current_measure, notes_to_add,
                                                   sounding_notes)

                    # Make new notes_to_add list of song_elements.Note objects.
                    notes_to_add = []

                    # Reset first-note-off flag.
                    first_note_off_at_tick = True

                if event_type == _EventType.NOTE_OFF:
                    # This is like a Measure start event, but we don't know how many notes
                    # will be turned off.
                    #   We need to record all note-off events, and then prepare a new note
                    # event with tie_from_previous set for all pitches.
                    if first_note_off_at_tick:
                        first_note_off_at_tick = False
                        # Add all currently sounding notes to notes_to_add, but then start removing
                        # them for each note_off. Added notes will tie-from-previous. This is just
                        # like a measure break.
                        for pitch in sounding_notes:
                            if sounding_notes[pitch]:
                                notes_to_add.append(Note(pitch, tie_from_previous=True))

                    # Remove note from notes_to_add if it exists (it might have been deleted
                    # already if there are multiple NOTE_ON MIDI events with a single NOTE_OFF);
                    # pretty midi will result in multiple NOTE_OFF events in this case at the
                    # same time.
                    for i, note_to_add in enumerate(notes_to_add):
                        if note_to_add.midi_pitch == note.pitch:
                            del notes_to_add[i]
                            break

                    sounding_notes[note.pitch] = False

                elif event_type == _EventType.MEASURE_START:
                    # Move to next measure.
                    measure_idx += 1
                    current_measure = track.new_measure(time_signatures[measure_idx])

                elif event_type == _EventType.NOTE_ON:
                    # New note.
                    # If note was already sounding, replace any existing tie_from_previous version.
                    remove_idx = None
                    for idx, added_note in enumerate(notes_to_add):
                        if added_note.midi_pitch == note.pitch:
                            remove_idx = idx
                    if remove_idx is not None:
                        del notes_to_add[remove_idx]
                    notes_to_add.append(Note(note.pitch, note.velocity))
                    sounding_notes[note.pitch] = True

                prev_tick = tick

            # Make sure no notes are still sounding.
            for sounding in sounding_notes.values():
                assert not sounding

            channel += 1
            if channel == _DRUM_CHANNEL:  # skip drum channel
                channel = _DRUM_CHANNEL + 1
            if channel > 15:
                print('Warning: too many tracks. Using channel %d for song %s' % (channel, name))

        return song



"""A collection of classes used to define a Song."""

from copy import deepcopy
from enum import Enum
from functools import total_ordering

from Core.exceptions import ArgumentError
import Core


class TimeSignature:
    """Represents a time signature."""
    def __init__(self, numerator=4, denominator=4):
        self.numerator = numerator
        self.denominator = denominator

    def __str__(self):
        return '%d/%d' % (self.numerator, self.denominator)

    def __eq__(self, ts2):
        return (isinstance(ts2, TimeSignature)
                and self.numerator == ts2.numerator
                and self.denominator == ts2.denominator)

    def __ne__(self, ts2):
        return not self == ts2


class Mode(Enum):
    """Enum to represent Major/Minor mode."""
    MAJOR = 0
    MINOR = 1


# Lookup tables mapping from pitch class (0-11) to string name of key.
_PC_TO_NAME_MAJOR = ('C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B')
_PC_TO_NAME_MINOR = ('C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'Bb', 'B')

# Lookup tables mapping from # sharps to pitch class and vice versa.
_SHARPS_TO_TONIC_MAJOR = {i: (7 * i) % 12 for i in range(-6, 7)}
_SHARPS_TO_TONIC_MINOR = {i: (7 * i - 3) % 12 for i in range(-6, 7)}
_TONIC_TO_SHARPS_MAJOR = {_SHARPS_TO_TONIC_MAJOR[i]: i for i in range(-6, 7)}
_TONIC_TO_SHARPS_MINOR = {_SHARPS_TO_TONIC_MINOR[i]: i for i in range(-6, 7)}

class Key:
    """Represents a musical key signature."""
    def __init__(self, tonic_pitch_class=None, num_sharps=None, mode=Mode.MAJOR):
        """Must specify either tonic_pitch_class or num_sharps.
        tonic_pitch_class: 0-11. 0=C, 1=C#, etc.
        num_sharps: positive number for # sharps. Negative number represents # of flats.
        mode: Mode enum. MAJOR or MINOR.
        """
        if ((tonic_pitch_class is None and num_sharps is None)
                or (tonic_pitch_class is not None and num_sharps is not None)):
            raise ArgumentError('Specify exactly one of {tonic_pitch_class, num_sharps}')

        assert isinstance(mode, Mode)
        self.mode = mode

        if tonic_pitch_class is not None:
            assert tonic_pitch_class >= 0 and tonic_pitch_class < 12
            self.tonic_pitch_class = tonic_pitch_class
        else:
            assert num_sharps > -7 and num_sharps < 7
            if mode == Mode.MAJOR:
                self.tonic_pitch_class = _SHARPS_TO_TONIC_MAJOR[num_sharps]
            else:
                self.tonic_pitch_class = _SHARPS_TO_TONIC_MINOR[num_sharps]

    def __str__(self):
        if self.mode == Mode.MAJOR:
            name_map = _PC_TO_NAME_MAJOR
        else:
            name_map = _PC_TO_NAME_MINOR
        return '%s %s' % (name_map[self.tonic_pitch_class], self.mode.name)

    def __eq__(self, key2):
        return (isinstance(key2, Key)
                and self.tonic_pitch_class == key2.tonic_pitch_class
                and self.mode == key2.mode)

    def number_of_sharps(self):
        """Returns the number of sharps in the key signature, or negative numbers for the number
        of flats, as in MIDI key signatures."""
        if self.mode == Mode.MAJOR:
            return _TONIC_TO_SHARPS_MAJOR[self.tonic_pitch_class]
        return _TONIC_TO_SHARPS_MINOR[self.tonic_pitch_class]



class TrackType(Enum):
    """Enum to represent JukeDeck-style type of a track (Melody/Harmony/etc.)"""
    UNKNOWN = 0
    MELODY = 1
    HARMONY = 2
    BASS = 3
    DRUMS = 4
    FX = 5


class Track:
    """Stores a single track (instrument) in a Song. Consists of multiple measures."""
    def __init__(self, name, program, channel, track_type, parent_song):
        self.name = name
        self.program = program  # MIDI program number
        self.channel = channel  # MIDI channel

        # Melody, harmony, bass, or drums.
        assert isinstance(track_type, TrackType)
        self.track_type = track_type

        assert isinstance(parent_song, Core.song.Song)
        self.parent_song = parent_song
        self.measures = []

    def __iter__(self):
        return iter(self.measures)

    def __eq__(self, t2):
        if (not (isinstance(t2, Track) and self.name == t2.name and self.program == t2.program
                 and self.channel == t2.channel and self.track_type == t2.track_type
                 and len(self.measures) == len(t2.measures))):
            return False
        for i, measure in enumerate(self.measures):
            if not measure == t2.measures[i]:
                return False
        return True

    def __str__(self):
        return '%s (program:%d, channel=%d, type=%s)' % (self.name, self.program, self.channel,
                                                         self.track_type.name)

    def new_measure(self, time_signature=None):
        """Adds a new measure at the end of the track and returns the measure."""
        next_tick = 0
        if self.measures:
            prev_measure = self.measures[-1]
            next_tick = prev_measure.start_tick + prev_measure.get_duration_ticks()
        measure = Measure(next_tick, self, time_signature)
        self.measures.append(measure)
        return measure


class Measure:
    """Stores a single measure of a single track. Consists of multiple Events."""
    def __init__(self, start_tick, parent_track, time_signature=None):
        assert isinstance(parent_track, Track)
        self.events = []
        self.start_tick = start_tick
        self.parent_track = parent_track
        if time_signature:
            self.time_signature = time_signature
        else:
            self.time_signature = parent_track.parent_song.time_signature
        assert isinstance(self.time_signature, TimeSignature)

    def __iter__(self):
        return iter(self.events)

    def __eq__(self, m2):
        if not (isinstance(m2, Measure)):
            return False

        # Ignore any final rests in a measure in comparison.
        events1 = deepcopy(self.events)
        events2 = deepcopy(m2.events)
        len_events1 = len(events1)
        for i in range(len(events1) - 1, -1, -1):
            if not events1[i].notes:
                len_events1 = i
            else:
                # As soon as we see notes, break.
                break

        len_events2 = len(events2)
        for i in range(len(events2) - 1, -1, -1):
            if not events2[i].notes:
                len_events2 = i
            else:
                # As soon as we see notes, break.
                break

        if len_events1 != len_events2:
            return False

        for i in range(0, len_events1):
            if not self.events[i] == m2.events[i]:
                return False
        return True

    def __str__(self):
        s = 'Measure:'
        for event in self:
            s += '\n\t%s' % str(event)
        return s

    def new_event(self, duration):
        event = Event(duration)
        self.events.append(event)
        return event

    def get_duration_ticks(self):
        """Returns the length of the measure in ticks, based on the time signature."""
        beats = self.time_signature.numerator * 4.0 / self.time_signature.denominator
        return int(beats * self.parent_track.parent_song.ticks_per_beat)


class Event:
    """Stores a single Event in a single track. This consists of all notes sounding at a particular
    moment in time, and a duration."""
    def __init__(self, duration=0):
        self.notes = []
        self.duration = int(duration)  # in ticks (as defined in the ancestor Song object)

    def __iter__(self):
        return iter(self.notes)

    def __eq__(self, e2):
        if not (isinstance(e2, Event) and self.duration == e2.duration
                and len(self.notes) == len(e2.notes)):
            return False

        # Don't require notes to be in same order. Treat as a set.
        for note in self.notes:
            for note2 in e2.notes:
                if note == note2:
                    break
            else:
                return False
        return True

    def __str__(self):
        s = 'Event duration: %d [' % self.duration
        for i, note in enumerate(self.notes):
            s += str(note)
            if i < len(self.notes) - 1:
                s += ', '
        s += ']'
        return s

    def append_note(self, note):
        self.notes.append(note)

    def new_note(self, midi_pitch, velocity=-1, tie_from_previous=False):
        note = Note(midi_pitch, velocity, tie_from_previous)
        self.notes.append(note)
        return note


@total_ordering
class Note:
    """Stores a single Note. This is either a note on event or a continuation of a previously
    sounding note. Note off events are handled implicitly by the presence of a later event where
    the note is not sounding."""
    def __init__(self, midi_pitch, velocity=-1, tie_from_previous=False):
        self.midi_pitch = midi_pitch
        self.velocity = velocity  # set to -1 if tie_from_previous==True
        self.tie_from_previous = tie_from_previous

    def __eq__(self, n2):
        return (isinstance(n2, Note) and self.midi_pitch == n2.midi_pitch
                and self.velocity == n2.velocity
                and self.tie_from_previous == n2.tie_from_previous)

    def __lt__(self, n2):
        assert isinstance(n2, Note)
        return self.midi_pitch < n2.midi_pitch

    def __str__(self):
        tie_string = ''
        if self.tie_from_previous:
            tie_string = 'TIE'
        return '%d[v=%d %s]' % (self.midi_pitch, self.velocity, tie_string)




"""Classes for representing a song in a MIDI-like format."""

# Requires the following pacakges:
#   * mido        -- for MIDI export
#   * pretty_midi -- for MIDI import
#
# Install:
#   pip install pretty_midi    (includes mido)
#
# Measures consist of events and durations. To make an event mid-measure, there must be a "rest"
# event to fill in the space before the event. A rest is simply an Event with an empty note
# collection.

import gzip
import pickle

from Core.song_elements import Track, TimeSignature, Key, TrackType
import Core


class Song:
    """Stores an entire symbolic score for a song. Contains multiple tracks."""
    def __init__(self, name='', ticks_per_beat=480): #TODO: add other params (midi_path, h5_path, original key?)
        self.version = '0.2'
        self.name = name
        self.time_signature = TimeSignature()
        self.key = Key(tonic_pitch_class=0)
        self.tracks = []
        #self.midi_path = midi_path
        # ticks_per_beat represents the finest granularity of time slices that we can represent.
        # If reading from MIDI, can be set to the same value as in the source MIDI file
        self.ticks_per_beat = ticks_per_beat

        self.time_signature_changes = []

        # TODO: represent mid-score key signature changes

    def __str__(self):
        return "Song: %s. %d track%s. Time signature: %s. Key: %s" % (
            self.name, len(self.tracks), '' if len(self.tracks) == 1 else 's',
            str(self.time_signature), str(self.key))

    def __iter__(self):
        return iter(self.tracks)

    def __eq__(self, s2):
        if not (isinstance(s2, Song) and
                self.name == s2.name and
                self.time_signature == s2.time_signature and
                self.key == s2.key and
                len(self.tracks) == len(s2.tracks) and
                self.ticks_per_beat == s2.ticks_per_beat):
            return False
        # Verify all tracks from self show up in s2. order does not matter; allow permutations.
        for track in self.tracks:
            for track2 in s2.tracks:
                if track == track2:
                    break
            else:
                return False
        return True

    def print_tracks(self, verbose=False, max_measures=4):
        """Prints list of all tracks. If verbose==True prints all measures and events."""
        for track in self:
            print(track)
            if verbose:
                for i, measure in enumerate(track.measures):
                    if i >= max_measures:
                        break
                    print(measure)
                print()

    def new_track(self, name, program, channel, track_type=TrackType.UNKNOWN):
        """Creates and appends a new Track to this song. Returns the new track."""
        track = Track(name, program, channel, track_type, self)
        self.tracks.append(track)
        return track

    def save(self, output_filename):
        """Save this Song object to a zipped pickle file."""
        with gzip.open(output_filename, 'wb') as f:
            pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)

    #def export_to_midi(self, output_filename):
    #    """Export this song to a MIDI file."""
    #    Core.song_midi_converter.SongMidiConverter.export_midi(self, output_filename)

    #@staticmethod
    #def create_song_from_midi(midi_filename):
    #    """Returns a new Song object improted from the given MIDI file."""
    #    return Core.song_midi_converter.SongMidiConverter.create_song_from_midi(midi_filename)

    @staticmethod
    def load(filename):
        """Load a Song object from a Song zipped pickle file. Returns the Song object."""
        with gzip.open(filename, 'rb') as f:
            return pickle.load(f)

    def export_vector(self):
        raise NotImplementedError()

    def import_vector(self):
        raise NotImplementedError()

    def adjust_ties_(self, clean_up, smooth_harmony):
        """Provides cleaning and smoothing options for ties."""
        for track in self:
            if smooth_harmony and track.track_type != TrackType.HARMONY:
                continue
            prev_sounding_notes = set()
            for measure in track:
                for event in measure:
                    sounding_notes = set()
                    for note in event:
                        if note.midi_pitch in prev_sounding_notes:
                            if smooth_harmony:
                                note.tie_from_previous = True
                        elif clean_up:
                            note.tie_from_previous = False
                        sounding_notes.add(note.midi_pitch)
                    prev_sounding_notes = sounding_notes

    def clean_ties(self):
        """Remove any tie_from_previous Note parameters where the note wasn't previously sounding."""
        self.adjust_ties_(clean_up=True, smooth_harmony=False)

    def force_smooth_harmony(self):
        """Add ties for any harmony tracks between repeated notes, so that the harmony part has
        minimal rearticulations."""
        self.adjust_ties_(clean_up=True, smooth_harmony=True)





import os
import unittest
import tempfile

from mido import MidiFile
from mido.messages import Message
from mido.midifiles.meta import MetaMessage
from pretty_midi import PrettyMIDI

from Core.pretty_midi_utils import is_monophonic, get_all_time_signatures
from Core.song import Song
from Core.song_elements import TrackType, TimeSignature, Key, Mode, Measure, Event, Note
from Core.song_midi_converter import SongMidiConverter

class Test_song(unittest.TestCase):
    @staticmethod
    def make_song_empty():
        s = Song('Test2')
        s.time_signature = TimeSignature(3, 8)
        s.key = Key(num_sharps=-2, mode=Mode.MINOR)
        s.new_track('Flute', 1, 1, TrackType.MELODY)
        return s

    @staticmethod
    def make_song_monophonic():
        s = Song('TestMonophonic')
        track = s.new_track('Flute', 1, 1, TrackType.MELODY)
        for j in range(7):
            measure = track.new_measure()
            for k in range(4):
                event = measure.new_event(k+1)
                if k % 3 == 0:
                    event.new_note(41 - j, 64)
        return s

    @staticmethod
    def make_song_polyphonic():
        return Test_song.make_song(1, 4)

    @staticmethod
    def make_song(num_tracks, num_measures, include_drums=False):
        s = Song('Random Song')
        for i in range(num_tracks):
            if include_drums and i == num_tracks-1:
                # Make a drum track.
                track = s.new_track('Drums', 0, 9, TrackType.DRUMS)
                for j in range(num_measures):
                    measure = track.new_measure()
                    for k in range(num_tracks):
                        event = measure.new_event(s.ticks_per_beat / (2**k))
                        for m in range(k):
                            if k % (1+m) == 0:
                                event.new_note(41 - m, 64 + 16 * m)
            else:
                # Make a pitched track.
                track = s.new_track('Track %d' % i, i, i)
                for j in range(num_measures):
                    measure = track.new_measure()
                    for k in range(i+1):
                        event = measure.new_event(s.ticks_per_beat / (2**k))
                        if k % 3 < 1:
                            event.new_note(72 - 3 * i, 64 + 16 * i)
                        if k % 3 < 2:
                            event.new_note(72 - 3 * i - 3 * (k+1), 64 + 16 * i)
        return s

    @staticmethod
    def make_song_with_ties():
        s = Song('Song with Ties')
        track = s.new_track('Track 0', 0, 0)

        for i in range(12):
            if i % 4 == 0:
                measure = track.new_measure()
            event = measure.new_event(s.ticks_per_beat)
            event.new_note(48 + 2*i, 64)
            if i > 0:
                event.new_note(48+2*(i-1), tie_from_previous=True)
            if i > 1:
                event.new_note(48+2*(i-2), tie_from_previous=True)

        return s

    @staticmethod
    def make_song_with_ties_2():
        s = Song('Song with Ties', ticks_per_beat=192)
        track = s.new_track('Track 0', 0, 0)

        measure = track.new_measure()
        event = measure.new_event(10)
        event = measure.new_event(185)
        event.new_note(33, 127)
        event = measure.new_event(1)
        event = measure.new_event(154)
        event.new_note(33, 127)
        event = measure.new_event(145)
        event.new_note(33, 127)
        event = measure.new_event(172)
        event.new_note(33, 127)
        event = measure.new_event(96)
        event.new_note(33, 127)
        event = measure.new_event(5)
        event.new_note(29, 127)

        measure = track.new_measure()
        event = measure.new_event(202)
        event.new_note(29, tie_from_previous=True)
        event = measure.new_event(146)
        event.new_note(29, 127)
        event = measure.new_event(140)
        event.new_note(31, 127)
        event = measure.new_event(1)
        event = measure.new_event(167)
        event.new_note(31, 127)
        event = measure.new_event(1)
        event = measure.new_event(83)
        event.new_note(31, 127)
        event = measure.new_event(28)
        event.new_note(33, 127)

        return s

    def test_song_str(self):
        song1 = Song('Test')
        self.assertEqual(str(song1), 'Song: Test. 0 tracks. Time signature: 4/4. Key: C MAJOR')

        song2 = self.make_song_empty()
        self.assertEqual(str(song2), 'Song: Test2. 1 track. Time signature: 3/8. Key: G MINOR')

    def test_song_save_load(self):
        """Make a Song and save to a temp file."""
        song1 = self.make_song_empty()
        filename = tempfile.mktemp()
        song1.save(filename)

        # Delete the original!
        del song1

        # Load, make sure it worked.
        song2 = Song.load(filename)

        # Clean up.
        os.remove(filename)

        self.assertEqual(str(song2), 'Song: Test2. 1 track. Time signature: 3/8. Key: G MINOR')

    def song_to_midi_to_song(self, song):
        """Make a song and export to midi."""
        filename = tempfile.mktemp() + '.mid'
        SongMidiConverter.export_midi(song, filename)

        # Read back from MIDI.
        song2 = SongMidiConverter.create_song_from_midi(filename)
        song2.name = song.name

        # Clean up.
        os.remove(filename)

        print ('imported song:')
        song2.print_tracks(True)

        # Compare the two versions.
        self.assertEqual(song, song2)

    def test_empty_song_to_midi_to_song(self):
        self.song_to_midi_to_song(Song())

    def test_song_to_midi_to_song_2_3(self):
        self.song_to_midi_to_song(self.make_song(2, 3, True))

    def test_song_to_midi_to_song_4_1(self):
        self.song_to_midi_to_song(self.make_song(4, 2, True))

    def test_song_to_midi_to_song_5_8(self):
        self.song_to_midi_to_song(self.make_song(5, 8, True))

    def test_song_to_midi_to_song_ties(self):
        self.song_to_midi_to_song(self.make_song_with_ties())

    def test_song_to_midi_to_song_ties_2(self):
        self.song_to_midi_to_song(self.make_song_with_ties_2())

    @staticmethod
    def create_midi_file():
        midifile = MidiFile(ticks_per_beat=192)
        track = midifile.add_track('main bass')
        track.append(Message('program_change', channel=0, program=3, time=0))
        track.append(MetaMessage('time_signature', numerator=4, denominator=4))

        track.append(Message('note_on', channel=0, note=33, velocity=127, time=6154))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=185))
        track.append(Message('note_on', channel=0, note=33, velocity=127, time=1))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=154))
        track.append(Message('note_on', channel=0, note=33, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=145))
        track.append(Message('note_on', channel=0, note=33, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=172))
        track.append(Message('note_on', channel=0, note=33, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=96))

        track.append(Message('note_on', channel=0, note=29, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=29, velocity=0, time=207))
        track.append(Message('note_on', channel=0, note=29, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=29, velocity=0, time=146))

        track.append(Message('note_on', channel=0, note=31, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=31, velocity=0, time=140))
        track.append(Message('note_on', channel=0, note=31, velocity=127, time=1))
        track.append(Message('note_off', channel=0, note=31, velocity=0, time=167))
        track.append(Message('note_on', channel=0, note=31, velocity=127, time=1))
        track.append(Message('note_off', channel=0, note=31, velocity=0, time=83))

        track.append(Message('note_on', channel=0, note=33, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=216))
        track.append(Message('note_on', channel=0, note=33, velocity=127, time=1))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=162))
        track.append(Message('note_on', channel=0, note=33, velocity=127, time=1))
        track.append(Message('note_off', channel=0, note=33, velocity=0, time=140))

        return midifile

    @staticmethod
    def create_midi_file_with_time_signatures():
        midifile = MidiFile(ticks_per_beat=4)  # 1 tick = 1/16th note
        track = midifile.add_track('main track')

        # 1 empty measure

        # Measure 2: half rest, then 1 quarter notes, then quarter rest
        track.append(Message('note_on', channel=0, note=60, velocity=127, time=16 + 8))
        track.append(Message('note_off', channel=0, note=60, velocity=0, time=4))

        # Measure 3: Time signature change to 3/4, then 3 quarter notes
        track.append(MetaMessage('time_signature', numerator=3, denominator=4, time=4))
        track.append(Message('note_on', channel=0, note=63, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=63, velocity=0, time=4))
        track.append(Message('note_on', channel=0, note=64, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=64, velocity=0, time=4))
        track.append(Message('note_on', channel=0, note=65, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=65, velocity=0, time=4))

        # Measure 4: Time signature change to 5/8, then quarter rest, then 3 eighth notes
        track.append(MetaMessage('time_signature', numerator=5, denominator=8, time=0))
        track.append(Message('note_on', channel=0, note=63, velocity=127, time=4))
        track.append(Message('note_off', channel=0, note=63, velocity=0, time=2))
        track.append(Message('note_on', channel=0, note=64, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=64, velocity=0, time=2))
        track.append(Message('note_on', channel=0, note=65, velocity=127, time=0))
        track.append(Message('note_off', channel=0, note=65, velocity=0, time=2))

        return midifile

    @staticmethod
    def create_song_with_ties():
        s = Song('Song with Ties', ticks_per_beat=4)
        track = s.new_track('Track 0', 0, 0)
        m = track.new_measure(TimeSignature())

        m.new_event(1)

        e = m.new_event(1)
        e.new_note(69, 100)

        m.new_event(1)

        e = m.new_event(1)
        e.new_note(76, 100)


        e = m.new_event(1)
        e.new_note(76, 100, True)

        e = m.new_event(1)
        e.new_note(74, 100)

        e = m.new_event(1)
        e.new_note(74, 100, True)

        e = m.new_event(1)
        e.new_note(72, 100)


        e = m.new_event(1)
        e.new_note(69, 100)

        m.new_event(1)

        e = m.new_event(1)
        e.new_note(76, 100)

        e = m.new_event(1)


        e = m.new_event(1)
        e.new_note(74, 100)

        m.new_event(1)

        e = m.new_event(1)
        e.new_note(72, 100)

        e = m.new_event(1)
        e.new_note(74, 100)

        #e = m.new_event(1)
        #e.new_note(72, 100)

        return s

    def test_midi_to_song_to_midi(self):
        # Create MIDI file object in memory.
        midifile = Test_song.create_midi_file()

        # Write to file.
        tmp_midi_orig = tempfile.mktemp() + '.mid'
        midifile.save(tmp_midi_orig)

        # Import to Song object.
        s = SongMidiConverter.create_song_from_midi(tmp_midi_orig)

        # Export to midi.
        tmp_midi_new = tempfile.mktemp() + '.mid'
        SongMidiConverter.export_midi(s, tmp_midi_new)

        # Compare original MIDI with exported MIDI.
        midifile_new = MidiFile(tmp_midi_new)

        self.assertEqual(len(midifile.tracks), len(midifile_new.tracks))
        new_track = midifile_new.tracks[0]
        for i, msg in enumerate(midifile.tracks[0]):
            self.assertEqual(str(msg), str(new_track[i]))

        # Clean up.
        os.remove(tmp_midi_orig)
        os.remove(tmp_midi_new)

    def test_instrument_is_monophonic(self):
        """Test is_monophonic function, True case."""
        # Make mono song, convert to midi, load in pretty_midi, extract instrument.
        s_mono = self.make_song_monophonic()
        filename = tempfile.mktemp() + '.mid'
        SongMidiConverter.export_midi(s_mono, filename)
        midi = PrettyMIDI(filename)
        instrument = midi.instruments[0]
        self.assertTrue(is_monophonic(instrument))

    def test_instrument_is_polyphonic(self):
        """Test is_monophonic function, False case."""
        # Make poly song, convert to midi, load in pretty_midi, extract instrument.
        s_poly = self.make_song_polyphonic()
        filename = tempfile.mktemp() + '.mid'
        SongMidiConverter.export_midi(s_poly, filename)
        midi = PrettyMIDI(filename)
        instrument = midi.instruments[0]
        self.assertFalse(is_monophonic(instrument))

    def verify_time_signatures(self, time_signatures):
        self.assertEqual(len(time_signatures), 4)

        self.assertEqual(time_signatures[0].numerator, 4)
        self.assertEqual(time_signatures[0].denominator, 4)

        self.assertEqual(time_signatures[1].numerator, 4)
        self.assertEqual(time_signatures[1].denominator, 4)

        self.assertEqual(time_signatures[2].numerator, 3)
        self.assertEqual(time_signatures[2].denominator, 4)

        self.assertEqual(time_signatures[3].numerator, 5)
        self.assertEqual(time_signatures[3].denominator, 8)

    def test_get_all_time_signatures(self):
        midifile = Test_song.create_midi_file_with_time_signatures()

        tmp_midi_orig = tempfile.mktemp() + '.mid'
        midifile.save(tmp_midi_orig)

        pm = PrettyMIDI(tmp_midi_orig)
        time_signatures = get_all_time_signatures(pm)
        os.remove(tmp_midi_orig)

        self.verify_time_signatures(time_signatures)

    def test_export_time_signatures_midi(self):
        # Create a midifile with time signature changes.
        midifile = Test_song.create_midi_file_with_time_signatures()

        tmp_midi_orig = tempfile.mktemp() + '.mid'
        midifile.save(tmp_midi_orig)

        # Load it into a Song object.
        s = SongMidiConverter.create_song_from_midi(tmp_midi_orig)

        # Export to midi.
        tmp_midi_new = tempfile.mktemp() + '.mid'
        SongMidiConverter.export_midi(s, tmp_midi_new)

        # Get time signatures from new MIDI file using PrettyMIDI.
        pm = PrettyMIDI(tmp_midi_new)
        time_signatures = get_all_time_signatures(pm)

        os.remove(tmp_midi_orig)

        # Verify time signatures.
        self.verify_time_signatures(time_signatures)

    def test_song_to_midi_to_song_ties_3(self):
        s = self.create_song_with_ties()
        s.print_tracks(True)
        self.song_to_midi_to_song(s)


if __name__ == '__main__':
    unittest.main()
